import java.io.*;
import java.util.*;

public class SolTI85String extends Object
{
    final byte LEVEL_VERSION         = 0;
    final byte SOL_USGARDID          = 6;
    final String STRING_DESCRIPTION  = "----- Solomon's Key Level File v1.0 -----";
    TI85String ti85String = null;

    public SolTI85String( String filename, String authorname, String levelname, byte[][] levels ) 
        throws Exception, IOException
    {
        ti85String = new TI85String( STRING_DESCRIPTION, filename );
        ti85String.writeToDisk( getStringData(authorname, levelname, levels) );
    }
    
    // gets a byte array representing the data portion of a TI85 string
    private byte[] getStringData(String author, String level, byte[][] levels) throws IOException
    {
        ByteArrayOutputStream stream = new ByteArrayOutputStream();

        stream.write( SOL_USGARDID );               // 1 byte for sol level ID
        stream.write( "SK".getBytes() );            // 2 bytes for "SK" to make sure sol's key level
        
        stream.write( ( (SolEditor.isCheatingEnabled) ? (byte) 0x80 : (byte)0x00 )
                      | LEVEL_VERSION );            // 1 byte for level version number
        stream.write( (byte)level.length() );       // 1 byte for level name length
        stream.write( level.getBytes() );           // level.length() bytes for level name
        stream.write( (byte)author.length() );      // 1 byte for author name
        stream.write( author.getBytes() );          // author.length() bytes for author name
        stream.write( (byte)levels.length );        // 1 byte for number of levels
        stream.write( compressLevels(levels) );     // N bytes for the compressed levels

        return ( stream.toByteArray() );
    }
    
    private byte[] compressLevels( byte[][] levels ) throws IOException
    {
        ByteArrayOutputStream stream = new ByteArrayOutputStream();

        for (int i=0; i<levels.length; i++)
        {
            stream.write( compressLevel(levels[i]) );
        }
        
        return ( stream.toByteArray() );
    }

    // Bytes are stored as follows for compressed data: 
    //  1 byte for number of bytes in level
    //  3 bytes for most common tiles (values 01b, 10b, and 11b)
    //  28 bytes for bitmap (2 bits for each tile in level)
    //  (# of Other Tiles) bytes for list of other tiles (values 00b in bitmap)
    private byte[] compressLevel( byte[] level ) throws IOException
    {
        ByteArrayOutputStream stream = new ByteArrayOutputStream();

        // find the 3 most frequent tiles in the level
        byte[] freqTiles = null;

        freqTiles = find3MostFrequentBytes( level );
        
        // 2 bits in bitmap for each tile (so 1/4 size of regular level)
        byte[] bitmap = new byte[SolEditor.LEVELSIZE * 2 / 8];

        byte[] masks = new byte[3];
        int bitmapIndex = 0;
        ByteArrayOutputStream otherTiles = new ByteArrayOutputStream();
        for (int i=0; i<SolEditor.LEVELSIZE; i++)
        {
            // every 4 rotates, need to set masks to 2 MSBs
            if ( (i%4) == 0 )
            {
                masks[0] = (byte) 0x40;        // value 1 in 2 most MSBs
                masks[1] = (byte) 0x80;        // value 2 in 2 most MSBs
                masks[2] = (byte) 0xC0;        // value 3 in 2 most MSBs
                bitmapIndex = i/4;
            }

            // put the most frequent tiles in the bitmap.
            //  if tile is not one of most freq., leave bits 0 in bitmap
            if ( level[i] == freqTiles[0] )
            {
                bitmap[bitmapIndex] |= masks[0];
            }
            else if ( level[i] == freqTiles[1] )
            {
                bitmap[bitmapIndex] |= masks[1];
            }
            else if ( level[i] == freqTiles[2] )
            {
                bitmap[bitmapIndex] |= masks[2];
            }
            else
            {
                // when it is a non frequent tile, write it to the stream
                otherTiles.write( level[i] );
            }

            masks[0] = (byte)((masks[0] & 0xFF) >> 2);
            masks[1] = (byte)((masks[1] & 0xFF) >> 2);
            masks[2] = (byte)((masks[2] & 0xFF) >> 2);
        }
        byte[] others = otherTiles.toByteArray();
        byte compLevelLength = (byte)( 1 +              // 1 byte for number of bytes in level
                                       3 +              // 3 bytes for most common tiles in level
                                       bitmap.length +  // 28 bytes for bitmap
                                       others.length ); // # of other tiles
        
        ByteArrayOutputStream compStream = new ByteArrayOutputStream();
        
        // write the number of bytes in level
        compStream.write(compLevelLength);
        
        // write 3 most frequent tiles
        compStream.write(freqTiles[0]);
        compStream.write(freqTiles[1]);
        compStream.write(freqTiles[2]);
        
        // write the bitmap
        compStream.write(bitmap);
        
        // write the other tiles
        compStream.write(others);
        
        return compStream.toByteArray();
    }
    
    private byte[] find3MostFrequentBytes( byte[] byteArray )
    {
        // new byte array to hold most frequent tiles
        byte[] freqTiles = new byte[3];
        
        // init array to hold count of each byte to 0
        int[] countArray = new int[256];
        
        // get the count
        for (int i=0; i<byteArray.length; i++)
            countArray[ byteArray[i] & 0xFF ]++;

        freqTiles[0] = findAndRemoveMostFrequentByte( countArray );
        freqTiles[1] = findAndRemoveMostFrequentByte( countArray );
        freqTiles[2] = findAndRemoveMostFrequentByte( countArray );
        
        return freqTiles;
    }
    
    private byte findAndRemoveMostFrequentByte( int[] countArray )
    {
        byte mostFreq = 0;
        int mostCount = 0;
        
        // see which byte occurs most often
        for (int i=0; i<countArray.length; i++)
        {
            if ( countArray[i] > mostCount )
            {
                mostFreq = (byte) i;
                mostCount = countArray[i];
            }
        }

        // set the count of this byte to 0
        countArray[mostFreq & 0xFF] = 0;

        return mostFreq;
    }
    
/*
    public static void main(String[] args)
    {
	    int[][] intData =
	    {
	      {
	        0x41,0x00,0x00,0x00,0x00,0x00,0x00,0x44,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	        0x00,0x00,0x00,0x00,0x82,0x82,0x82,0x82,0x00,0x80,0x82,0x44,0x00,0x00,0x00,0x42,
	        0x00,0x00,0x00,0x80,0x41,0x43,0x82,0x42,0x82,0x00,0x00,0x82,0x00,0x00,0x00,0x00,
	        0x00,0x00,0x82,0x42,0x44,0x82,0x00,0x87,0x00,0x82,0x44,0x00,0x82,0x00,0x44,0x00,
	        0x80,0x82,0x82,0x09,0x82,0x42,0x00,0x85,0x00,0x00,0x82,0x00,0x0A,0x82,0x80,0x80,
	        0x00,0x00,0x00,0x82,0x00,0x00,0x00,0x83,0x00,0x00,0x00,0x82,0x82,0x00,0x00,0x00,
	        0x40,0x00,0x00,0x00,0x00,0x00,0x81,0x82,0x84,0x00,0x00,0x00,0x00,0x00,0x00,0x00 
          },
          {
	        0x00,0x00,0x00,0x00,0x00,0x45,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	        0x00,0x00,0x00,0x00,0x00,0x80,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	        0x00,0x00,0x00,0x00,0x46,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	        0x00,0x00,0x00,0x00,0x80,0x00,0x00,0x85,0x00,0x00,0x82,0x00,0x09,0x82,0x80,0x80,
	        0x00,0x00,0x00,0x0a,0x00,0x00,0x05,0x04,0x04,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	        0x40,0x00,0x00,0x04,0x05,0x04,0x81,0x82,0x04,0x00,0x00,0x00,0x00,0x00,0x00,0x00
	      }
	    };


        byte[][] data = new byte[intData.length][];
        for (int i=0; i<data.length; i++)
        {
            data[i] = new byte[intData[i].length];
            for (int j=0; j<data[i].length; j++)
                data[i][j] = (byte) intData[i][j];
        }

        try
        {
            SolTI85String str = new SolTI85String("test.85g","Mike P","New Levels", data);
        }
        catch (Exception e)
        {
            System.out.println(e);
        }

        try { System.in.read(); } catch (Exception e) {}

    }
*/    
}