
import java.io.*;

/**
 * This class is for a ti85 string.
 */

public class TI85String extends Object
{
    private String filename = null;
    private String description = null;

    /**
     * Class Constructor.
     * @param description Description used for this 85s string file.
     * @param filename the new 85s file name (with path, if necessary)
     * @param data the data to be entered into the 85s string file.
     * @exception IOException
     * @exception Exception if description > 0x29 characters
     * @exception Exception if filename (w/out path and extension) > 8 characters
     */
    public TI85String(String description, String filename)
    {
        this.filename = filename;
        this.description = description;
    }
    
    /**
     * Write the data to the file specified.
     */
    public void writeToDisk( byte[] data ) throws IOException, Exception
    {
        File file = new File(filename);
        
        FileOutputStream fstream = new FileOutputStream( file );

        byte[] header = getHeader( description );
        byte[] entry = getEntry( parseFilename(filename), data );
        byte[] bytesToEOF = getBytesToEOF( entry );
        byte[] checksum = getChecksum( entry );

        fstream.write( header );
        fstream.write( bytesToEOF );
        fstream.write( entry );
        fstream.write( checksum );
        
        fstream.close();
    }

    
    // determine the name of the file without any directories or extension (.85s)
    private String parseFilename(String filename) throws Exception
    {
        String fname = null;
        
        // index of path separator
        int index = filename.lastIndexOf(System.getProperty("file.separator"));

        if ( index == -1 )
            fname = filename;           // didn't find, so filename in working dir
        else
            fname = filename.substring(index + 1);  // fname is after file separator char

        // index of . in extension
        index = fname.lastIndexOf('.');
        if ( index != -1 )
            fname = fname.substring(0, index);

        if ( fname.length() > 8 )
            throw new Exception("filename (without path and extension) cannot be greater " +
                                "than 8 characters");
            
        return fname;
    }
    
    private byte[] getHeader(String desc) throws IOException, Exception
    {
        ByteArrayOutputStream stream = new ByteArrayOutputStream();

        // signature
        stream.write( "**TI85**".getBytes() );
        
        // eof marker and further signature
        stream.write( (byte) 0x1A );
        stream.write( (byte) 0x0C );
        stream.write( (byte) 0x00 );
        
        // cannot be too large
        if ( desc.length() > 0x29 )
            throw new Exception("String header description " +
                                "cannot be greater than " +
                                0x29 +
                                " bytes");

        // pad with Zero if not 0x29 characters
        while (true)
        {
            if ( desc.length() < 0x29 )
                desc += '\0';
            else
                break;
        }
        // now write this string
        stream.write( desc.getBytes() );
        stream.write( new Character('\0').toString().getBytes() );
        
        return stream.toByteArray();
    }

    private byte[] getBytesToEOF(byte[] entry)
    {
        byte[] length = new byte[2];
        
        length[0] = (byte)(entry.length & 0xFF);
        length[1] = (byte)(entry.length >> 8);
        
        return length;
    }
    
    private byte[] getEntry(String stringName, byte[] data) throws IOException
    {
        ByteArrayOutputStream stream = new ByteArrayOutputStream();
        byte dataByte = 0;

        // num of bytes to end of string name (little endian)
        dataByte = (byte) ((2 + 1 + 1 + stringName.length()) & 0xFF);
        stream.write( dataByte );
        dataByte = (byte) ((2 + 1 + 1 + stringName.length()) >> 8);
        stream.write( dataByte );

	    // load string data size + 2 bytes (little endian)
	    dataByte = (byte) ((data.length + 2) & 0xFF);
	    stream.write( dataByte );
	    dataByte = (byte) ((data.length + 2) >> 8);
	    stream.write( dataByte );

        // "string" type of variable on calc
        stream.write( (byte) 0x0C );

	    // length of ti85 string's name
	    stream.write( (byte) stringName.length() );
	    
	    // now write the string name
	    stream.write( stringName.getBytes() );
	    
	    // load data size + 2 bytes (little endian) - again
	    dataByte = (byte) ((data.length + 2) & 0xFF);
	    stream.write( dataByte );
	    dataByte = (byte) ((data.length + 2) >> 8);
	    stream.write( dataByte );
	    
	    // load data size (no adding 2 bytes)
	    dataByte = (byte) (data.length & 0xFF);
	    stream.write( dataByte );
	    dataByte = (byte) (data.length >> 8);
	    stream.write( dataByte );

	    // write the data
	    stream.write( data );

        // get the byte array up to this point
        byte []byteArray = stream.toByteArray();

        return stream.toByteArray();
    }

    private byte[] getChecksum(byte []data)
    {
    	int sum=0;

	    for (int i=0; i<data.length; i++)
		    sum += data[i] & 0xFF;

	    return ( new byte[] { (byte)(sum & 0xFF), (byte)((sum >> 8) & 0xFF) } );
    }
    
}